from .geometry import Point, Convex
from .pod import Podable, Pod

class Agent(Podable):

    # Initialization Destruction:
    def __init__( self, identifier= 0, group=0, position= Point(0.0, 0.0), shape= None):
        self._id= identifier
        self._group= group
        self._tile= 0
        self._center= Point( position.x(), position.y() )
        self._shape= shape
        if self._shape is None :
            self._shape= Convex().initializeSquare(0.4)
        self._matter= 10+group

    # Accessor: 
    def group(self):
        return self._group
    
    def id(self):
        return self._id
    
    def tile(self):
        return self._tile
    
    def matter(self):
        return self._matter
    
    def position(self):
        return self._center
    
    def shape(self):
        return self._shape
    
    def body(self):
        return self._shape.copy(self._center)
    
    # Convex accessor : 
    def box(self):
        return self.body().box()
    
    def radius(self):
        r= 0.0
        zero= Point()
        for p in self._shape._points :
            d= zero.distance( p )
            r= max( d, r )
        return r

    # Construction:
    def setId(self, aInteger):
        self._id= aInteger
        return self
    
    def setGroup(self, aInteger):
        self._group= aInteger
        return self
    
    def setTile(self, aInteger):
        self._tile= aInteger
        return self
    
    def setPosition(self, aPoint):
        self._center= aPoint
        return self
    
    def setMatter(self, aInteger):
        self._matter= aInteger
        return self
    
    def setShape(self, aConvex):
        self._shape= aConvex
        return self
    
    def setShapeRegular(self, size= 1.0):
        self._shape.initializeSquare(size)
        return self

    def setShapeRegular(self, numberOfVertex= 6, size= 1.0):
        self._shape.initializeRegular( numberOfVertex, size)
        return self

    def setBody(self, aConvex):
        self._center= aConvex.center()
        subCenter= Point( -self._center.x(), -self._center.y() )
        self._shape= aConvex.copy(subCenter) 
        return self
    
    # Pod interface:
    def asPod(self):
        return Pod().fromLists( 
            ["Agent"], 
            [self.id(), self.group(), self.matter(), self.tile()],
            self.position().asList(),
            [ self.shape().asPod() ]
        )
    
    def fromPod( self, aPod ):
        integers= aPod.integers()
        self.setId( integers[0] )
        self.setGroup( integers[1] )
        self.setMatter( integers[2] )
        self.setTile( integers[3] )
        self.setPosition( Point().fromList( aPod.values() ) )
        self.setShape( Convex().fromPod( aPod.children()[0] ) )
        return self
    
    # str:
    def str(self, typeName= "Agent"): 
        if self.group() :
            return typeName + f"-{self.group()}.{self.id()} {self.box()}"
        return typeName + f"-{self.id()} {self.box()}"
        
    def __str__(self):
        return self.str()


import math

class Point():
    # Initialization Destruction:
    def __init__( self, x= 0.0, y=0.0 ):
        self._x= float(x)
        self._y= float(y)

    def fromPoints(self, p1, p2):
        self._x= p2._x - p1._x
        self._y= p2._y - p1._y
        return self
    
    # Accessors
    def x(self):
        return self._x
    
    def y(self):
        return self._y
    
    # Transform :
    def asTuple(self): 
        return (self._x, self._y)
    
    def asList(self): 
        return [self._x, self._y]
    
    # Construction
    def setx(self, value):
        self._x= value
        return self
    
    def sety(self, value):
        self._y= value
        return self

    def set( self, x, y ):
        return self.setx(x).sety(y)
    
    def round( self, precision=0 ):
        self._x= round( self._x, precision )
        self._y= round( self._y, precision )
        return self
    
    def fromList( self, aList ):
        return self.setx( aList[0] ).sety( aList[1] )
    
    def fromTrigo( self, aRadAngle, aDistance= 1.0 ):
        self._x= math.cos( aRadAngle ) * aDistance
        self._y= math.sin( aRadAngle ) * aDistance
        return self

    # Object operator:
    def copy(self):
        cpy= type(self)()
        cpy._x= self.x()
        cpy._y= self.y()
        return cpy
    
    # Operator: 
    def add(self, another):
        self._x+=another._x
        self._y+=another._y
        return self

    def __add__(self, another):
        return Point( self._x+another._x,  self._y+another._y )

    def __sub__(self, another):
        return Point( self._x-another._x,  self._y-another._y )

    #Comparison:
    def __eq__(self, another):
        return ( self.x() == another.x()
            and self.y() == another.y()
        )

    def distanceSquare(self, another):
        delta= another - self
        dx, dy = delta.asTuple()
        return dx*dx + dy*dy
     
    def distance(self, another):
        delta= another - self
        dx, dy = delta.asTuple()
        return math.sqrt( dx*dx + dy*dy )
    
    # Trigo
    def norm2(self):
        return self._x*self._x + self._y*self._y
    
    def length(self):
        return math.sqrt( self.norm2() )

    def crossProduct(self, another):
        return (self._x * another._y - self._y * another._x)
    
    def dotProduct(self, another):
        return (self._x * another._x + self._y * another._y)
    
    def angleAbs(self, another):
        dot = self.dotProduct(another)
        n1 = self.length()
        n2 = another.length()
        cos_theta = dot / (n1 * n2)
        cos_theta = max( min(cos_theta, 1.0), -1.0)
        return  math.acos(cos_theta)
    
    def angle(self, another):
        dot = self.dotProduct(another)
        cross = self.crossProduct(another)
        angle = math.atan2(cross, dot)  # retourne [-pi, pi]
        return angle

        dot = self.dotProduct(another)
        n1 = self.length()
        n2 = another.length()
        cos_theta = dot / (n1 * n2)
        cos_theta = max( min(cos_theta, 1.0), -1.0)
        return  math.acos(cos_theta)
    
    def angleClockwise(self, another):
        a= self.angle(another)
        if a == 0.0 :
            return a
        if a < 0.0 :
            return -a
        return 2*math.pi-a
    
    # Collision
    def isColliding(self, point, distance=0.001):
        between= point - self
        return ( between.norm2() < (distance * distance) )

    def isCollidingLine(self, a, b, distance=0.001):
        vab= b-a
        v1= a-self
        v2= b - self
        return ( v1.length() + v2.length() ) < vab.length() + distance

    # list of points:
    def isClockwise(self, another):
        return self.crossProduct(another) < 0.0
    
    def isCounterClockwise(self, another):
        return self.crossProduct(another) > 0.0
    
    def sortRadial( self, points ):
        size= len(points)
        
        # Find point minimal distance to the left-bottom corner :
        corner= Point(
            min( [ p.x() for p in points ] ),
            min( [ p.y() for p in points ] )
        )
        minDist2= corner.distanceSquare( points[0] )
        for i in range(1, size) :
            dist2= corner.distanceSquare( points[i] )
            if dist2 < minDist2 :
                p= points[0]
                points[0]= points[i]
                points[i]= p
                minDist2= dist2

        # compute angles :
        ref= points[0]-self
        angles= [ ref.angleClockwise( p-self ) for p in points ]
        
        # Sort points by angles
        notok= True
        while notok :
            notok= False
            for i in range(size-2, 0, -1) :
                if angles[i] > angles[i+1] :
                    a= angles[i+1]
                    p= points[i+1]
                    angles[i+1]= angles[i]
                    points[i+1]= points[i]
                    angles[i]= a
                    points[i]= p
                    notok= True
        return points

    # to str
    def str(self): 
        # Myself :
        s= f"({round(self.x(), 2)}, {round(self.y(), 2)})"
        return s
    
    def __str__(self): 
        return self.str()

class Line :
    # Initialization Destruction:
    def __init__( self, a= Point(), b= Point()):
        self._p1= a
        self._p2= b

    # Accessors
    def point1(self):
        return self._p1

    def point2(self):
        return self._p2
    
    def vector(self):
        return self._p2 - self._p1
    
    def lenght(self):
        return self.vector().length()
    
    # Trigo
    def determinant(self):
        return (self.point1()._x * self.point2()._y
                - self.point1()._y * self.point2()._x)
    
    def denominator(self):
        dx = self._p2._x - self._p1._x
        dy = self._p2._y - self._p1._y
        return dx*dx + dy*dy

    def projectionPoint(self, point):
        dx = self._p2._x - self._p1._x
        dy = self._p2._y - self._p1._y
        denom= dx*dx + dy*dy

        if denom == 0:
            return self._p1.copy()

        t = ((point._x - self._p1._x) * dx + (point._y - self._p1._y) * dy)
        t/= denom
        t = max(0.0, min(1.0, t))

        return Point(self._p1._x + t*dx, self._p1._y + t*dy)
    
    def distancePoint(self, p):
        line= Line( p, self.projectionPoint(p) )
        return line.lenght()

    # Collision
    def isColliding( self, another, collide= Point(), distance=0.001 ):
        x_diff = Point( self.point1()._x - self.point2()._x, another.point1()._x - another.point2()._x)
        y_diff = Point( self.point1()._y - self.point2()._y, another.point1()._y - another.point2()._y)

        divisor = Line(x_diff, y_diff).determinant()
        if divisor == 0:
            return False

        d = Point( self.determinant(), another.determinant() )
        collide._x = Line(d, x_diff).determinant() / divisor
        collide._y = Line(d, y_diff).determinant() / divisor
        
        return (
            collide.isCollidingLine( self.point1(), self.point2(), distance )
            and collide.isCollidingLine( another.point1(), another.point2(), distance )
        )
    
    # to str
    def str(self): 
        # Myself :
        s= f"{self.point1()}->{self.point2()}"
        return s
    
    def __str__(self): 
        return self.str()


import math

from .basic import Point, Line
from .box import Box
from .mesh import Mesh
from .convex import Convex
from .grid import Grid

#ToDo: from .plane import Plane

clockAngles= [ 0.0, 2.0*math.pi/6.0, math.pi/6.0, 0.0,
               math.pi/-6.0, -2.0*math.pi/6.0, math.pi/-2.0,
               -2.0*math.pi/3.0, -5.0*math.pi/6.0, math.pi,
               5.0*math.pi/6.0, 2.0*math.pi/3.0, math.pi/2.0
               ]

clockPositions= [
    Point().fromTrigo(x) for x in clockAngles
]
clockLenght= 13



"""
Pick'n Del - an HackaGame game 
"""

from . import carrier, master, world, player

# Game Component:
Carrier= carrier.Carrier
World= world.World

# Hackagame Game:
GameEngine= master.GameEngine
GameMaster= master.GameMaster

# Players
BasicBot= player.BasicBot
#BlindBot= player.BlindBot
#ShellPlayer= player.ShellPlayer




from .carrier import Carrier
from ... import scene, Tile, artist
from ...geometry import Point, Convex, Box
import hacka as hk
import random

class Mission:
    def __init__( self, start= 0, final= 0, reward= 0, owner= 0 ):
        self.start= start
        self.final= final
        self.reward= reward
        self.owner= owner

    def fromList(self, aList):
        self.start= aList[0]
        self.final= aList[1]
        self.reward= aList[2]
        self.owner= aList[3]
        return self

    def asList(self):
        return [self.start, self.final, self.reward, self.owner]
    
    def asTuple(self):
        return self.start, self.final, self.reward, self.owner

class World( scene.Scene ):
    def __init__(self, name="Pick'nDel", numberOfPlayers= 1):
        super().__init__()
        self.setAgentFactory(Carrier)
        self._name= name
        self._missions= []
        self._encumbers= []
        # Initialize Artist :
        self._artist= artist.Artist().initializePNG( "shot-pickndel.png", 800, 600 )
        self._artist.flip()
        self._artist.fitBox( Box([Point(-0.5, -0.5), Point(9.5, 6.5)] ), 10 )
        #self._artist.fitBox( self.box(), 10 )
        self.marketBrush= self._artist._panel[6]
        self.marketBrush.width= 8

    # Accessor: 
    def name(self):
        return self._name
    
    def carrierTile(self, iCarrier= 1, iPlayer= 1):
        return self.agent(iCarrier, iPlayer).tile()
    
    def carrierMission(self, iCarrier= 1, iPlayer= 1):
        mission= self.agent(iCarrier, iPlayer).mission()
        if mission != 0 : 
            return mission
        return self.missionIndexes()[0]

    def carrierGoal(self, iCarrier= 1, iPlayer= 1):
        mission= self.agent(iCarrier, iPlayer).mission()
        if mission != 0 : 
            return self.mission( mission ).final
        mission= self.missionIndexes()[0]
        return self.mission( mission ).start
    
    def missions(self):
        return self._missions

    def isMission(self, iMission):
        return ( 0 < iMission and iMission <= len(self._missions) )

    def mission(self, index):
        return self._missions[index-1]
    
    def missionIndexes(self):
        l= []
        i= 1
        for m in self._missions :
            if m.start > 0 :
                l.append(i)
            i+= 1
        return l
    
    def carrierTiles(self, iPlayer):
        return [ m.tile() for m in self.agents(iPlayer) ]
    
    def encumber(self, iTile):
        return self._encumbers[iTile-1]

    # Construction:
    def initializeMoves(self):
        for group in range( self.numberOfGroups() ) :
            for car in self.agents(group) :
                car.setMove(0)

    def append( self, tile, encumber= 0.0 ):
        super(World, self).append(tile)
        self._encumbers.append(encumber)

    def setEncumber( self, iTile, value ):
        self._encumbers[iTile-1]= value
        return self

    def resetEncumbers( self, defaultValue= 0.0 ):
        self._encumbers= [ defaultValue for i in range(self.size()) ]
        return self
    
    def initializeLine( self, size, tileSize= 1.0, separation= 0.1, connect=True ):
        super(World, self).initializeLine(size, tileSize, separation, connect)
        self.resetEncumbers()
        return self
    
    def initializeGrid( self, matrix, tileSize= 1.0, separation=0.1, encumbers= [[],[]], connect=True ):
        super(World, self).initializeGrid(matrix, tileSize, separation, connect)
        self.resetEncumbers()
        for iTile, enc in zip( encumbers[0], encumbers[1] ) :
            self.setEncumber( iTile, enc )
        return self

    def clear( self ):
        super(World, self).clear()
        self._encumbers= []

    def addTile( self, aTile, encumber= 0.0 ):
        super(World, self).addTile(aTile)
        self._encumbers.append(encumber)
        return self._size

    # Mission :
    def setMissions( self, aListOfTuples, pay= 124 ):
        self._missions= [
            Mission(iFrom, iTo, pay, 0)
            for iFrom, iTo in aListOfTuples 
        ]
    
    def clearMissions(self):
        self._missions= []
        for group in range(self.numberOfGroups()+1) :
            for iCarrier in range(1, self.numberOfAgents(group)+1) :
                self.agent( iCarrier, group ).setMission(0)
        return self

    def addMission( self, iFrom, iTo, pay= 124 ):
        self._missions.append( Mission(iFrom, iTo, pay, 0) )
        return len(self._missions)
    
    def addMissionAtRandom( self ):
        tileIndexes= range( 1, self.size()+1 )
        return self.addMission( random.choice( tileIndexes ), random.choice(tileIndexes) )

    def updateMission(self, iMission, iFrom, iTo, pay, owner):
        self._missions[iMission-1]= Mission(iFrom, iTo, pay, owner)

    def addRandomMission(self):
        bound= self._engine._map.size()+1
        iFrom= random.randrange(1, bound)
        iTo= random.randrange(1, bound)
        pay= 10+random.randrange(bound)
        self._engine.addMission( iFrom, iTo, pay )
        return iFrom, iTo, pay

    # Moving:
    def move(self, iFrom, clockDir):
        #print( f">>> move {iFrom}, {clockDir} ({self.encumber(iFrom)})" )
        if self.tile(iFrom).count() == 0 or clockDir == 0 :
            return iFrom
        if random.random() < self.encumber(iFrom) :
            return iFrom
        iTo= self.clockposition( iFrom, clockDir ) 
        return self.teleport(iFrom, iTo)

    def teleport( self, iFrom, iTo ):
        if self.tile(iFrom).count() == 0 or self.tile(iTo).count() :
            return False
        # move:
        # Get from iFrom
        carrier= self.tile(iFrom).agent()
        self.tile(iFrom).clear()

        # Set on iTo
        self.tile(iTo).append(carrier)
        carrier.setTile( iTo )
        carrier.setPosition( self.tile(iTo).position() )
        return iTo
    
    # Podable:
    def asPod( self ):
        return hk.Pod( self._name, [], [], [ super(World, self).asPod(), self.missionsAsPod() ] )
    
    def missionsAsPod(self):
        missionPod= hk.Pod( "Missions" )
        for m in self._missions :
            missionPod.append( hk.Pod( "Mission", m.asList() ) )
        return missionPod
    
    def carriersAsPod(self):
        podMobiles= hk.Pod( "Carriers" )
        for group in range( self.numberOfGroups() ):
            for car in self.agents( group ):
                mPod= hk.Pod( "carrier", [group, car.id(), car.tile(), car.mission()] )
                podMobiles.append(mPod)
        return podMobiles

    def fromPod( self, aPod ):
        self._name= aPod.label()
        super(World, self).fromPod( aPod.child(1) )
        self.missionsFromPod(  aPod.child(2) )
        return self
    
    def missionsFromPod(self, aPod):
        self._missions= []
        for childPod in aPod.children() :
            self._missions.append( Mission().fromList( childPod.integers() ) )
        return self._missions
    
    def carriersFromPod(self, aPod):
        self.clearAgents()
        for pod in aPod.children() :
            iPlayer= pod.integer(1)
            iCarrier= pod.integer(2)
            pos= pod.integer(3)
            mis= pod.integer(4)
            carrier= self.popAgentOn( pos, iPlayer )
            assert carrier.id() == iCarrier
            carrier.setMission(mis)
        return self._agents

    def setOnPodState(self, aPod):
        self.missionsFromPod( aPod.child(1) )
        self.carriersFromPod( aPod.child(2) )
        return aPod.integer(1)
    
    # Rendering :
    def render(self):
        self._artist.drawScene( self )
        # Market:
        self._artist.drawPolygon(
            [6.55, 6.55, 9.5, 9.5], [2.45, -0.6, -0.6, 2.45],
            self.marketBrush
        )
        self._artist._fontSize= 20
        self._artist.write( 6.6, 2.2, "Market Place:", self.marketBrush )
        self._artist._fontSize= 16
        sep= 0.0
        for i in self.missionIndexes() :
            mFrom, mTo, pay, iPlayer= self.mission(i).asTuple()
            self._artist.write( 6.8, 1.9-sep, f".{i}", self.marketBrush) 
            self._artist.write( 7.2, 1.9-sep, f"- {mFrom} to: {mTo}", self.marketBrush )
            if iPlayer == 0 :
                self._artist.write( 8.5, 1.9-sep, f"({pay} ¢)", self.marketBrush )
            else :
                self._artist.write( 8.4, 1.9-sep, f"(Team-{iPlayer})", self.marketBrush )
            sep+= 0.24
        # Finalize:
        self._artist.flip()




import math
from ..pod import Podable, Pod
from .basic import Point, Line
from .convex import Point, Convex
from .mesh import Mesh
#from .box import Box

class Grid() : # ToDo: Podable
    # Initialization
    def __init__(self, values= [[0]], bottomleft= Point(0.0, 0.0), resolution=0.1):
        self.initialize(values)
        self._bottomleft= bottomleft
        self._resolution= resolution

    # Construction
    def initialize(self, aGrid):
        self._grid= aGrid
        self._height= len(self._grid)
        self._width= len(self._grid[0])
        return self

    def setCell( self, x, y, state ):
        i, j= self.inTable(x, y)
        self._grid[i][j]= state
        return self
    
    # Accessors
    def values(self):
        return self._grid
    
    def height(self):
        return self._height

    def width(self):
        return self._width

    def dimention(self):
        return ( self._width, self._height )
    
    def inTable(self, x, y):
        return self._height-y, x-1

    def cell(self, x, y):
        assert 0 < x and x <= self.width() and 0 < y and y <= self.height()
        i, j = self.inTable(x, y)
        return self._grid[i][j]

    def bottomleft(self):
        return self._bottomleft

    def resolution(self):
        return self._resolution

    def point(self, x, y):
        #return Point(x, y)
        p= self._bottomleft
        r= self._resolution
        return Point( p.x() + x*r, p.y() + y*r )

    def valueMinMax(self):
        if self.height() == 0 :
            return (0, 0)
        minMatter= -1
        maxMatter= self.cell(1, 1)

        for ix in range(1, self.width()+1):
            for iy in range(1, self.height()+1):
                matter= self.cell(ix, iy)
                if minMatter == -1 :
                    minMatter= matter
                if matter >= 0 :
                    minMatter= min(minMatter, matter)
                    maxMatter= max(maxMatter, matter)
        return (minMatter, maxMatter)
    
    # Tests
    def isCell(self, x, y, state):
        return ( self.cell(x, y) == state )
    
    # Filling :
    def fill(self, ix, iy, matter):
        w, h= self.dimention()
        old= self.cell(ix, iy)
        if old == matter :
            return Point(ix, iy)
        toFill= [(ix, iy)]
        cumul= Point()
        count= 0
        while len(toFill) > 0 :
            for cx, cy in toFill :
                self.setCell(cx, cy, matter)
                cumul.add( Point(cx, cy) )
                count+= 1

            nextToFill= []
            for cx, cy in toFill :
                for nx, ny in [(cx-1, cy), (cx+1, cy), (cx, cy-1), (cx, cy+1)] :
                    if 0 < nx and nx <= w and 0 < ny and ny <= h and self.cell(nx, ny) == old :
                        nextToFill.append( (nx, ny) )
            toFill= nextToFill
        return Point( cumul.x()/count, cumul.y()/count )

    def fillMasks(self, masks, ix, iy, iCluster, radius):
        w, h= self.dimention()
        ref= self.cell(ix, iy)
        masks.setCell(ix, iy, iCluster)
        toFill= [(ix, iy)]
        cumul= Point()
        count= 0
        diameter= radius+radius
        it= 0
        while it <= diameter and len(toFill) > 0 :
            it+=1
            nextToFill= []
            for cx, cy in toFill :
                cumul.add( Point(cx, cy) )
                count+= 1
                for nx, ny in [(cx-1, cy), (cx+1, cy), (cx, cy-1), (cx, cy+1)] :
                    if (0 < nx and nx <= w and 0 < ny and ny <= h
                        and masks.cell(nx, ny) != iCluster and self.cell(nx, ny) == ref ):
                        masks.setCell(nx, ny, iCluster)
                        nextToFill.append( (nx, ny) )
            toFill= nextToFill
        return Point( cumul.x()/count, cumul.y()/count )

    def filter(self, value, modif):
        for i in range( self.height() ) :
            for j in range( self.width() ) :
                if self._grid[i][j] == value :
                    self._grid[i][j]= modif

    # Clustering :
    def clustering(self, matter, radius):
        marks, means= self.clusterInit(matter, radius)
        minDistance= 1.0
        while means.size() > 0 and minDistance > 0.001 :
            marks, newMeans= self.clusterIterate(matter, means, radius)
            for p, np in zip( means.points(), newMeans.points() ) :
                minDistance= min(minDistance, p.distance(np))
            means= newMeans
        return marks, means
    
    def clusterInitOld(self, matter, radius):
        means= self.clusterInitMeans(radius)
        marks, newMeans= self.clusterIterate(matter, means, radius)
        # Search free cluster...
        w, h= self.dimention()
        for x in range(1, w+1) :
            for y in range(1, h+1) :
                if marks.cell(x, y) == 0 and self.cell(x, y) == matter :
                    iCluster= means.size()+1
                    center= self.fillMasks( marks, x, y, iCluster, radius )
                    means.append(center)
        return marks, means
    

    def clusterInit(self, matter, radius):
        means= self.clusterInitMeans(radius)
        w, h= self.dimention()

        # Initialize marks on 0 :
        marks= Grid(
            [ [ 0 for j in range(w)]  for i in range(h) ],
            self.bottomleft(), self.resolution()
        )
        
        # Color the marks :
        maxRadius= int(radius*1.6)
        gridMeans= []
        count= 0
        for i in range(means.size()) :
            x, y= self.localPointToCoordinate( means.point(i+1) )
            x, y, dist= self.searchClosest(matter, x, y)
            if dist <= radius :
                count+= 1
                center= self.fillMasks( marks, x, y, count, maxRadius )
                gridMeans.append( center )
    
        # Search free cluster...
        w, h= self.dimention()
        for x in range(1, w+1) :
            for y in range(1, h+1) :
                if marks.cell(x, y) == 0 and self.cell(x, y) == matter :
                    iCluster= means.size()+1
                    center= self.fillMasks( marks, x, y, iCluster, maxRadius )
                    means.append(center)
        
        return marks, means
    
    def clusterInitMeans(self, radius):
        cosPi6= 0.86602540378
        distance= radius*2
        pi6Distance= cosPi6*distance
        w, h= self.dimention()
        nbWidth= int(max(w//(distance), 1))
        nbHeight= int(max(h//(pi6Distance), 1))
        means= []
        wMarge= w - nbWidth*distance + radius
        hMarge= h - nbHeight*pi6Distance + radius
        for i in range(nbHeight) :
            marge= wMarge
            lineSize= nbWidth
            if i % 2 ==1 :
                marge+= radius
                lineSize-= 1
            for j in range(lineSize):
                means.append( Point(marge+j*distance, hMarge+i*pi6Distance) )
        return Mesh( means )
    
    def localPointToCoordinate(self, aPoint):
        return (
            int( min( self.width(), max( 1, int(round(aPoint.x())) ))),
            int( min( self.height(), max( 1, int(round(aPoint.y())) )))
        )
    
    def clusterIterate(self, matter, proposedMeans, radius):
        w, h = self.dimention()

        # Initialize toVisit on mean-positions :
        toVisits= []
        myMeans= Mesh()
        count= 0
        for i in range(proposedMeans.size()) :
            x, y= self.localPointToCoordinate( proposedMeans.point(i+1) )
            x, y, dist= self.searchClosest(matter, x, y)
            if dist <= radius :
                toVisits.append( [(x, y)] )
                myMeans.append( proposedMeans.point(i+1) )
                count+= 1
        nbCluster= myMeans.size()

        # Initialize marks on 0 :
        marks= Grid(
            [ [ 0 for j in range(w)]  for i in range(h) ],
            self.bottomleft(), self.resolution()
        )

        # Initialize newMeans :
        clusterSum= [ Point() for i in range(nbCluster) ]
        clusterCount= [ 0 for i in range(nbCluster) ]

        while count > 0 :
            visited= []
            # Mark
            for i in range(nbCluster) :
                iCl= i+1
                for x, y in toVisits[i] :
                    p= Point(x, y)
                    cCl= marks.cell(x, y)
                    if cCl == 0 :
                        marks.setCell( x, y, iCl )
                        visited.append( (x, y) )
                    elif myMeans.point(iCl).distanceSquare(p) < myMeans.point(cCl).distanceSquare(p) :
                        marks.setCell( x, y, iCl )

            ## Enlarge :
            count= 0
            toVisits= [ [] for i in range(nbCluster) ]
            for x, y in visited :
                i= marks.cell(x, y)-1
                clusterSum[i].add( Point(x, y) )
                clusterCount[i]+= 1
                for cx, cy in [ (x+1, y), (x-1, y), (x, y+1), (x, y-1) ] :
                    if 0 < cx and cx <= w and 0 < cy and cy <= h and self.cell(cx, cy) == matter :
                        toVisits[i].append( (cx, cy) )
                        count+=1
        return marks, Mesh( [ Point( p.x()/c, p.y()/c ) for p, c in zip(clusterSum, clusterCount) ] )
    
    def clusterApplyBirdDistance(self, matter, means):
        w, h = self.dimention()
        clusterCenters= [ Point() for i in range( means.size() ) ]
        clusterCount= [ 0 for i in range( means.size() ) ]
        mask= Grid( [ [ 0 for j in range(w)]  for i in range(h) ] )
        for x in range( 1, w+1 ):
            for y in range( 1, h+1 ):
                if self.cell(x, y) == matter :
                    p= Point(x, y)
                    iCluster= means.searchClosestTo( p )
                    mask.setCell(x, y, iCluster)
                    clusterCenters[iCluster-1].add(p)
                    clusterCount[iCluster-1]+= 1
                    #cluster= 
        for i in range( means.size() ) :
            clusterCenters[i]._x/= clusterCount[i]
            clusterCenters[i]._y/= clusterCount[i]
        
        dists= [ clusterCenters[i].distance( means.point(i+1) ) for i in range(means.size()) ]
        return mask, Mesh( clusterCenters ), dists

    # Cuting :
    def cutingRectangles(self, state, expectedLenght):
        pos= self.searchLine(state)
        rectangles= []
        cellsSize= int( expectedLenght/self.resolution() )
        
        while pos :
            x, y= pos
            rect= self.maxRectangle(x, y, cellsSize)
            self.setRectangleOn( rect, -1 )
            rectangles.append( rect )
            pos= self.searchLine(state)
        
        for rect in rectangles :
            self.setRectangleOn(rect, state)
        return rectangles
    
    def maxRectangle(self, x, y, expectedCellsLenght):
        state= self.cell(x, y)
        width, height= self.dimention()
        x2= x+1
        y2= y+1
        xOk, yOk= True, True
        while xOk or yOk :
            xOk= xOk and x2 <= width and (x2-x) < expectedCellsLenght
            # Increase on x:
            iy= y
            while xOk and iy < y2 :
                xOk= xOk and (self.cell( x2, iy ) == state)
                iy+= 1
            if xOk :
                x2+= 1
            # Increase on y:
            yOk= yOk and y2 <= height and (y2-y) < expectedCellsLenght
            ix= x
            while yOk and ix < x2 :
                yOk= yOk and (self.cell( ix, y2 ) == state)
                ix+= 1
            if yOk :
                y2+= 1
        # Then return
        return [x, y, x2-1, y2-1]

    def searchClosest(self, matter, x, y):
        if self.cell(x, y) == matter :
            return (x, y, 0)
        w, h= self.dimention()
        for distance in range(1, max( w, h) ) :
            for d1 in range(distance):
                d2= distance-d1
                for cx, cy in [(x+d2, y+d1), (x-d1, y+d2), (x-d2, y-d1), (x+d1, y-d2) ] :
                    if 0 < cx and cx <= w and 0 < cy and cy <= h and self.cell(cx, cy) == matter :
                        return (cx, cy, distance)
        return False, False, max( w, h)

    def searchLine(self, state=0):
        x, y= 1, 1
        width, height= self.dimention()
        while y <= height and self.cell(x, y) != state :
            x+= 1
            if x > width :
                x=1
                y+=1
        if y > height :
            return False
        return (x, y)
    
    def setRectangleOn(self, rect, state ):
        bx1, by1, bx2, by2= tuple(rect)
        for ix in range( bx1, bx2+1 ):
            for iy in range( by1, by2+1 ):
                self.setCell(ix, iy, state)
        return self

    # Shaping
    def computeFrontiers(self):
        frontiers= []
        frontSize= 0
        w, h = self.dimention()
        w1, h1= w+1, h+1
        # horizontal border :
        for x in range(1, w):
            frontSize= self.updateFrontiers(frontiers, frontSize, x, 1)
            frontSize= self.updateFrontiers(frontiers, frontSize, w1-x, h)
        # vertical border :
        for y in range(2, h):
            frontSize= self.updateFrontiers(frontiers, frontSize, 1, y)
            frontSize= self.updateFrontiers(frontiers, frontSize, w, h1-y)
        # middle cells :
        for x in range(2, w):
            for y in range(2, h):
                m= self.cell(x, y)
                if ( self.cell(x-1, y) != m or self.cell(x+1, y) != m 
                        or self.cell(x, y-1) != m or self.cell(x, y+1) != m ) :
                    frontSize= self.updateFrontiers(frontiers, frontSize, x, y)
        return frontiers
    
    def updateFrontiers(self, frontiers, frontSize, x, y ):
        matter= self.cell(x, y)
        if matter < 0 :
            return frontSize
        if matter >= frontSize :
            frontiers+= [ [] for i in range( frontSize, matter+1 ) ]
            frontSize= matter+1
            assert len(frontiers) == frontSize
        frontiers[matter].append( self.point(x, y) )
        return frontSize

    def makeConvexes(self, matter, radius):
        marks, means= self.clustering(matter, radius)
        frontiers= marks.computeFrontiers()
        convs= []
        epsilon= self.resolution()*0.001
        for points in frontiers[1:] :
            c= Convex( points )
            c.simplify( epsilon )
            convs.append(c)
        return convs

    def makeRectangles(self, state, expectedSize=1.0):
        rectangles= self.cutingRectangles(state, expectedSize)
        shapes= [ self.rectangleToConvex(rect) for rect in rectangles ]
        return shapes
    
    def rectangleToConvex(self, rect):
        r= self.resolution()
        e= r*0.1
        ox, oy= self.bottomleft().asTuple()
        bx1, by1, bx2, by2= tuple(rect)
        sx1= ox + (bx1-1)*r + e
        sy1= oy + (by1-1)*r + e
        sx2= ox + bx2*r - e
        sy2= oy + by2*r - e
        return Convex().fromZipped( [(sx1, sy1), (sx1, sy2), (sx2, sy2), (sx2, sy1)] )

    # to str
    def str(self, typeName="Grid"): 
        # Myself :
        let= ['·'] + [ str(i) for i in range(1, 10) ] + [ chr(65+i) for i in range(26) ]
        s= f"{typeName} {self.width()}x{self.height()}\n"
        for line in self._grid :
            s+= "| " + " ".join( [let[state] for state in line] ) +"\n"
        s+= "0 " + "-".join( ["-" for _ in range(self.width())] )
        return s
    
    def __str__(self):
        return self.str()


"""
Test - MoveIt Robot Class
"""
from ...geometry import Point, Convex
from ...agent import Agent

mobileConvex= Convex().initializeRegular(0.4, 8)

class Carrier(Agent):
    def __init__( self, identifier=0, owner=1, position= Point(0.0, 0.0), mission= 0):
        super().__init__(identifier, owner, position, mobileConvex)
        self._mission= mission
        self._clockMove= 0

    # Accessor:     
    def mission(self):
        return self._mission
    
    def setMission(self, iMission):
        self._mission= iMission

    def move(self):
        return self._clockMove

    def setMove(self, clockDir):
        self._clockMove= clockDir

    # Accessor: 
    def str(self, typeName= "Carrier"): 
        s= super(Carrier, self).str(typeName)
        s+= f" |{self._clockMove}, {self._mission}|"
        return s


import math

class Mesh():
    # Constructions :
    def __init__(self, pointsList= []):
        self._points= [ p for p in pointsList ]
    
    def copy(self): 
        cpy= Mesh()
        cpy._points= [ p.copy() for p in self._points ]
        return cpy
    
    # Accessors :
    def size(self):
        return len(self._points)

    def points(self):
        return self._points
    
    def point(self, i):
        return self._points[i-1]
    
    # Construction :
    def append(self, p):
        self._points.append(p)

    # Morphing :
    def asTuples(self):
        return [ p.asTuple() for p in self._points ]
    
    # Opperations :
    def round(self, floatDeep):
        for p in self._points :
            p.round(2)
        return self
    
    def searchClosestTo( self, position ):
        if self.size() == 0 :
            return 0
        d2min= position.distanceSquare( self.point(1) )
        imin= 1
        for i in range( 2, self.size()+1 ):
            d2= position.distanceSquare( self.point(i) )
            if d2 < d2min :
                imin= i
                d2min= d2
        return imin



import math
from ..pod import Podable, Pod
from .basic import Point, Line
from .box import Box

class Convex(Podable):

    # Constructor/Destructor:
    def __init__( self, aListOfPoints= [] ):
        self.initialize( aListOfPoints )
    
    def copy(self, position= Point(0.0, 0.0) ):
        cpy= type(self)()
        cpy._points= [ position + p.copy() for p in self.points() ]
        return cpy
    
    # Initialization:
    def initialize(self, aListOfPoints):
        self._points= aListOfPoints
        self.makeConvex()
        return self
    
    def initializeSquare(self, size):
        demi= size*0.5
        self._points= [
            Point( -demi, -demi ),
            Point( -demi, +demi ),
            Point( +demi, +demi ),
            Point( +demi, -demi )
        ]
        return self

    def initializeRegular(self, size, numberOfVertex= 6):
        radius= size*0.5
        self._points= []
        delta= math.pi/(numberOfVertex/2)
        angle= math.pi  - delta/2
        delta= math.pi/(numberOfVertex/2)
        for i in range(numberOfVertex) :
            p= Point( math.cos(angle)*radius, math.sin(angle)*radius)
            self._points.append(p)
            angle+= -delta
        self._points= self.center().sortRadial( self._points )
        return self
    
    def forcePoints(self, points):
        self._points= points
        return self
    
    # Accessor:
    def size(self):
        return len(self._points)
    
    def points(self):
        return self._points
    
    def box(self):
        if self.size() == 0 :
            return Box()
        return Box(self.points())
    
    # Derivated Attributs:
    def center(self):
        box= self.box()
        return box.center()

    # Morphing:
    def asList(self):
        l= []
        for p in self._points:
            l+= [p.x(), p.y()]
        return l
    
    def asLists(self, dx=0.0, dy=0.0):
        return [p.x()+dx for p in self._points], [p.y()+dy for p in self._points]
    
    def fromLists(self, listX, listY):
        self.initialize( [ Point(x, y) for x, y in zip(listX, listY) ] )
        return self

    def asZipped(self):
        return [ (p.x(), p.y()) for p in self._points ]

    def fromZipped( self, zipedList ):
        self.initialize( [ Point(x, y) for x, y in zipedList ] )
        return self

    # Construction:
    def round(self, precision):
        for p in self._points :
            p.round(precision)
        return self
    
    def setOnCenter(self):
        position= self.center()
        for p in self._points :
            p.set( p._x-position._x, p._y-position._y,  )
        return position

    # Convex
    def makeConvex(self):
        removed= []

        # Safe stop :
        if self.size() == 0 :
            # print( f"makeConvex. Empty" )
            return removed

        # print( f"makeConvex..." )

        # Sort points arround a center :
        center= self.center()
        self._points= center.sortRadial( self._points )

        # print( f"points: {', '.join( [str(x) for x in self._points]) } Center({center})" )

        # Build convex list :
        convex= [ self._points[0] ]
        i= 0
        for p in self._points[1:] + [self._points[0]] :
            # print( f"convex: {', '.join( [str(x) for x in convex]) } ({i})" )
            # Init:
            # print( f"test: {p}" )

            while (i > 0) and (convex[i] - convex[i-1]).isCounterClockwise( p - convex[i-1] ) :
                removed.append( 
                    convex.pop(i)
                )
                i-= 1
                # print( f"remove: {[ p.asTuple() for p in removed ]} ({i})" )
            # push
            convex.append(p)
            i+= 1
        
        #self._points= [ convex[-1] ] + convex[0:-1]
        # print( f"final convex: {', '.join( [str(x) for x in convex]) }" )
        self._points= convex[:-1]
        return removed

    # Collision:   
    def isIncludingPoint(self, aPoint):
        # Suppose self is convex...
        if self.size() < 3 :
            return False
        p1= self._points[-1]
        for p2 in self._points :
            if (p1-p2).crossProduct( aPoint-p1 ) < 0.0 :
                return False
            p1= p2
        return True
    
    def isIncludingLine(self, aLine):
        return (
            self.isIncludingPoint( aLine.point1() ) 
            and self.isIncludingPoint( aLine.point2() )
        )

    def isCollidingLine(self, aLine):
        if self.isIncludingPoint( aLine.point1() ) or self.isIncludingPoint( aLine.point2() ) :
            return True
        p1= self._points[-1]
        for p2 in self._points :
            if aLine.isColliding( Line(p1, p2) ) :
                return True
            p1= p2
        return False
    
    def isColliding( self, another ):
        # Induce a collision at box granulatrity
        if not self.box().isColliding( another.box() ) :
            return False
        # If a point of one of them is included in the other
        for p in self._points :
            if another.isIncludingPoint(p) :
                return True
        for p in another._points :
            if self.isIncludingPoint(p) :
                return True
        # If lines collides
        p1= self._points[-1]
        for p2 in self._points :
            if another.isCollidingLine( Line(p1, p2) ) :
                return True
            p1= p2
        return False

    # Distances:
    def distancePoint(self, aPoint):
        minLine= Line( self._points[-1], self._points[0] )
        minDist= minLine.distancePoint(aPoint)
        p1= self._points[0]
        for p2 in self._points[1:] :
            line= Line( p1, p2 )
            dist= line.distancePoint(aPoint)
            if dist < minDist :
                minDist= dist
                minLine= line
            p1= p2
        self._trace= minLine
        return minDist

    def distanceLine(self, aLine):
        dist1= self.distancePoint( aLine.point1() )
        line1= self._trace
        minDist= self.distancePoint( aLine.point2() )
        
        if dist1 < minDist :
            minDist= dist1
            self._trace= line1
        
        minDist= min(
            minDist,
            aLine.distancePoint( self._trace.point1() ),
            aLine.distancePoint( self._trace.point2() )
        )
        
        return minDist

    def distance(self, another):
        if self.isColliding(another):
            return 0.0
        p1= self._points[-1]
        p2= self._points[0]
        l12= Line(p1, p2)
        minDist= another.distanceLine(l12)
        trace= [l12, another._trace]
        
        p1= p2
        for p2 in self ._points[1:] :
            l12= Line(p1, p2)
            dist= another.distanceLine(l12)
            if dist < minDist :
                minDist= dist
                trace= [l12, another._trace]
            p1= p2
        self._trace= trace
        return minDist

    # Geometry operation
    def merge( self, another ):
        self._points+= another.points()
        return self.makeConvex()
    
    def simplify(self, epsilon):
        removed= []
        notok= True
        while notok and self.size() > 3 :
            notok= False
            i = self.size()-1
            i1= i-1
            for i2 in range( self.size() ) :
                line= Line( self._points[i1], self._points[i2] )
                if line.distancePoint( self._points[i] ) < epsilon :
                    notok= True
                    removed.append( self._points.pop(i) )
                    break
                i1= i
                i= i2
        return removed

    # to str
    def str(self, typeName="Convex"): 
        # Myself :
        s= f"{typeName} {len(self._points)}" 
        s+= str( [(round(x, 2), round(y, 2)) for x, y in self.box().asZip()] )
        return s
    
    def __str__(self): 
        return self.str()
    
    # Pod interface:
    def asPod(self):
        return Pod().fromLists( ["Convex"], [], self.asList(), [] )
    
    def fromPod( self, aPod ):
        values= aPod.values()
        self._points= [
            Point(x, y)
            for x, y in zip( values[::2], values[1::2] )
        ]
        return self



import math
from . import geometry
from .pod import Pod
from .geometry import Point, Convex
from .agent import Agent

class Tile(Agent):

    def __init__( self, identifier= 0, position= Point(0.0, 0.0), shape= None, matter=0):
        shape
        if shape is None :
            shape= Convex().initializeSquare(1.0)
        self._adjacencies= []
        self._agents= []
        super(Tile, self).__init__(identifier, 0, position, shape)
        self._matter= matter
        
    # Accessor:
    def adjacencies(self):
        return self._adjacencies

    def agents(self) :
        return self._agents
    
    def count(self) :
        return len( self._agents)
    
    def agent(self, i=1) :
        return self._agents[i-1]
    
    # Construction:
    def setAdjacencies( self, aList ):
        self._adjacencies= aList
        return self
    
    # Connection:
    def isConnecting(self, iTile):
        return (iTile in self.adjacencies())
    
    def connect(self, iTo):
        if iTo not in self._adjacencies :
            self._adjacencies.append(iTo)
            self._adjacencies.sort()
        return self

    def disconnect(self, iTo):
        self._adjacencies.remove(iTo)
        return self

    def connectAll( self, aList ):
        for iTo in aList :
            self.connect( iTo )
        return self
    
    def clockDirection( self, aPosition ):
        center= self.position()
        radius= self.radius()
        if center.distance( aPosition ) < radius :
            return 0
        clock= 1
        distance= aPosition.distance( center + geometry.clockPositions[1] )
        for i in range( 2, geometry.clockLenght ):
            option= center + geometry.clockPositions[i]
            d= aPosition.distance( option )
            if d < distance :
                clock= i
                distance= d
        return clock

    # Agent managment
    def append(self, aPod, brushId=0, shapeId=0 ): 
        self._agents.append( aPod )
        return self
    
    def clear(self):
        self._agents = []
        return self
    
    # Comparison :
    def centerDistance(self, another):
        return self.position().distance( another.position() )

    def bodyDistance(self, another):
        return self.body().distance( another.body() )

    # Pod interface:
    def asPod(self):
        return Pod().fromLists(   
            ["Tile"], 
            [self.id(), self.matter()] + self.adjacencies(),
            self.position().asList(),
            [self.shape().asPod()] + [ ag.asPod() for ag in self.agents() ]
        )
    
    def fromPod( self, aPod, agentFactory=Agent ):
        integers= aPod.integers()
        children= aPod.children()
        self.setId( integers[0] )
        self.setMatter( integers[1] )
        self.setAdjacencies( integers[2:] )
        self.setPosition( Point().fromList( aPod.values() ) )
        self.setShape( Convex().fromPod( aPod.children()[0] ) )
        self.clear()
        for podBod in children[1:] :
            self.append( agentFactory().fromPod( podBod ) )
        return self
    
    # Classical Class
    def copy(self):
        cpy= type(self)()
        return cpy.fromPod( self.asPod() )
    
    # to str
    def str(self, typeName="Tile"): 
        # Myself :
        s= super(Tile, self).str(typeName)
        s+= " matter-"+ str(self._matter)
        s+= " adjs"+ str(self._adjacencies)
        s+= f" agents({ len(self.agents()) })"
        return s
    
    def __str__(self): 
        return self.str()
    


from .color import percentColor
from .support import AbsSupport

import math, cairo
#from .geometry import Coord2, Segment

class SupportPNG( AbsSupport ):
    def __init__(self, width= 800, height= 600, filePath= "shot-hacka.png" ):
        self._canvas= cairo.ImageSurface( cairo.Format.RGB24, width, height )
        self._width= width
        self._height= height
        self._filePath= filePath
        self.save( self._filePath )
    
    # Accessor: 
    def width(self):
        return self._width
    
    def height(self):
        return self._height
    
    def canvas(self):
        return self._canvas
    
    def filePath(self):
        return self._filePath

    # Control:
    def clear(self):
        self._canvas= cairo.ImageSurface( cairo.Format.RGB24, self._width, self._height )
    
    def render(self):
        return self._canvas
    
    def flip(self):
        if self._filePath :
            arts= self.save( self._filePath )
        else :
            arts= self.render()
        self.clear()
        return arts

    def save( self, filePath ):
        self._canvas.write_to_png( filePath )
        return

    # Drawing primitives:
    def traceLine( self, pixxA, pixyA, pixxB, pixyB, strokeColor, strokeWidth ):
        ctx = cairo.Context( self._canvas )
        r, g, b= percentColor(strokeColor)
        ctx.set_line_width(strokeWidth)
        ctx.move_to(pixxA, pixyA)
        ctx.line_to(pixxB, pixyB)
        ctx.set_source_rgb( r, g, b )
        ctx.stroke()
        return self
    
    def traceCircle( self, pixx, pixy, radius, strokeColor, strokeWidth ):
        ctx = cairo.Context( self._canvas )
        r, g, b= percentColor( strokeColor )
        ctx.set_line_width( strokeWidth )
        ctx.arc(pixx, pixy, radius, 0, 2.0*math.pi)
        ctx.set_source_rgb( r, g, b )
        ctx.stroke()
        return self

    def fillCircle( self, pixx, pixy, radius, fillColor ):
        ctx = cairo.Context( self._canvas )
        r, g, b= percentColor( fillColor )
        ctx.arc(pixx, pixy, radius, 0, 2.0*math.pi)
        ctx.set_source_rgb( r, g, b )
        ctx.fill()
        return self

    def drawCircle( self, pixx, pixy, radius, fillColor, strokeColor, strokeWidth):
        ctx = cairo.Context( self._canvas )
        sr, sg, sb= percentColor(strokeColor)
        fr, fg, fb= percentColor( fillColor )
        ctx.set_line_width(strokeWidth)
        ctx.arc(pixx, pixy, radius, 0, 2.0*math.pi)
        ctx.set_source_rgb( fr, fg, fb )
        ctx.fill_preserve()
        ctx.set_source_rgb( sr, sg, sb )
        ctx.stroke()
        return self
    
    def tracePolygon( self, pixXs, pixYs, strokeColor, strokeWidth ):
        ctx = cairo.Context( self._canvas )
        r, g, b= percentColor(strokeColor)
        ctx.set_line_width(strokeWidth)
        if len(pixXs) > 0 :
            ctx.move_to( pixXs[0], pixYs[0] )
            for pixx, pixy in zip( pixXs[1:], pixYs[1:] ) :
                ctx.line_to(pixx, pixy)
            ctx.close_path()
        ctx.set_source_rgb( r, g, b )
        ctx.stroke()
        return self

    def fillPolygon( self, pixXs, pixYs, fillColor ):
        ctx = cairo.Context( self._canvas )
        r, g, b= percentColor( fillColor )
        if len(pixXs) > 0 :
            ctx.move_to( pixXs[0], pixYs[0] )
            for pixx, pixy in zip( pixXs[1:], pixYs[1:] ) :
                ctx.line_to( pixx, pixy )
            ctx.close_path()
        ctx.set_source_rgb( r, g, b )
        ctx.fill()
        return self

    def drawPolygon( self, pixXs, pixYs, fillColor, strokeColor, strokeWidth ):
        ctx = cairo.Context( self._canvas )
        sr, sg, sb= percentColor(strokeColor)
        fr, fg, fb= percentColor(fillColor)
        ctx.set_line_width(strokeWidth)
        if len(pixXs) > 0 :
            ctx.move_to( pixXs[0], pixYs[0] )
            for pixx, pixy in zip( pixXs[1:], pixYs[1:] ) :
                ctx.line_to(pixx, pixy)
            ctx.close_path()
        ctx.set_source_rgb( fr, fg, fb )
        ctx.fill_preserve()
        ctx.set_source_rgb( sr, sg, sb )
        ctx.stroke()
        return self


    # Writting primitives:
    def write( self, pixX, pixY, text, color, fontSize ):
        ctx = cairo.Context( self._canvas )
        r, g, b= percentColor(color)
        ctx.set_font_size(fontSize)
        ctx.set_source_rgb( r, g, b )
        ctx.move_to( pixX, pixY )
        ctx.show_text(text)
        ctx.stroke()
        return self



from .pod import Podable, Pod
from .geometry import Point, Box, Convex
from .tile import Tile
from .agent import Agent

import math

class Scene(Podable):

    # Constructor:
    def clear( self ):
        self._tiles= []
        self._agents= [[]]
        self._size= 0
        return self
    
    def __init__(self, shapes= [], epsilon= 0.01):
        self._factory= Agent #lambda identifier, group : Agent( identifier, group, shape=Convex().initializeRegular(0.8, 5) ).setMatter(1)
        self.clear()
        for s in shapes :
            self.createTile( s )
        self._epsilon= epsilon
    
    # Initialization:
    def fromShapes(self, shapes, matter):
        # Clean basis:
        self.clear()
        for s in shapes :
            self.createTile(s, matter)
    
    def fromGridConvexes(self, aGrid, tileSize=1.0, matters= False):
        self.clear()
        seam= aGrid.resolution() * 1.001
        self._epsilon= aGrid.resolution() * 0.001
        gridConvexRadius= max(2, round( (tileSize/seam)/2.0 ))

        if not matters :
            minMatter, maxMatter= aGrid.valueMinMax()
            matters= range(minMatter, maxMatter+1)
        
        # Foreach value possibility:
        for matter in matters :
            convexes= aGrid.makeConvexes(matter, gridConvexRadius)
            for conv in convexes :
                self.createTile(conv, matter)

        # Connect all elements:
        self.connectAllClose( seam )

        return self

    def fromGridRectangles(self, aGrid, tileSize= 1.0):
        self.clear()
        self._epsilon= aGrid.resolution() * 0.4

        # Foreach value possibility:
        minMatter, maxMatter= aGrid.valueMinMax()
        i= 0
        for matter in range( minMatter, maxMatter+1 ):
            # Add all shapes
            shapes= aGrid.makeRectangles(matter, tileSize)
            for s in shapes :
                i+= 1
                assert self.createTile(s, matter) == i
        
        # Connect all elements:
        self.connectAllClose( aGrid.resolution() )

        # Optimize the definition:
        for factor in [0.2, 0.4, 0.6, 0.8] :
            self.mergeAllPossible( aGrid.resolution() * factor, tileSize)

        return self
    
    # Accessor:
    def epsilon(self):
        return self._epsilon
    
    def size(self):
        return self._size
    
    def tiles(self):
        return self._tiles

    def tile(self, iCell):
        return self._tiles[iCell-1]

    def agent(self, iAgent, group=0 ):
        return self._agents[group][iAgent-1]

    def numberOfGroups(self):
        return len(self._agents)
    
    def numberOfAgents(self, group=0):
        if group < len(self._agents) :
            return len( self._agents[group] )
        return 0
    
    # Graph:
    def adjacencies(self, iTile) :
        return self.tile(iTile).adjacencies()
    
    def edges(self):
        edgeList= []
        for t in self.tiles() :
            edgeList+= [ (t.id(), neibor) for neibor in t.adjacencies() ]
        return edgeList

    def neighbours(self, iTile):
        neibs= [] 
        for iNei in self.adjacencies(iTile) :
            clockdir= self.tile(iTile).clockDirection( self.tile(iNei).position() )
            neibs.append( (iNei, clockdir) )
        return neibs
        
    def directions(self, iTile) : 
        cx, cy= self.tile(iTile).position().asTuple()
        neibor= self.adjacencies(iTile)
        positions= [ self.tile(i).position().asTuple() for i in neibor ]
        return [ (x-cx, y-cy) for x, y in positions ]
    
    def clockBearing(self, iTile):
        clock= [
            [ 0,  9,  0],
            [ 6,  0, 12],
            [ 0,  3,  0]
        ]
        positions= [ (int(round(x, 0)), int(round(y, 0))) for x, y in self.directions(iTile) ]
        return [ clock[1+x][1+y] for x, y in positions ]

    def completeClock(self, iTile):
        clock= [ iTile for i in range(13) ]
        for it, ic in self.neighbours(iTile) :
            clock[ic]= it
        return clock

    def clockposition(self, iTile, clockDir):
        return self.completeClock(iTile)[clockDir]

    # Test:
    def isAgent(self, iAgent, group= 1):
        return group < len(self._agents)  and iAgent-1 < len(self._agents[group])
    
    def isTile(self, iTile):
        return 0 < iTile and iTile <= self.size()
    
    def isEdge(self, iFrom, iTo):
        return iTo in self.tile(iFrom).adjacencies()
    
    def box(self):
        if self._size == 0 :
            return Box()
        box= self.tile(1).box()
        for t in self.tiles()[1:] :
            box.merge( t.box() )
        return box

    def testNumberOfAgents(self):
        nb1= 0
        nb2= 0
        for t in self.tiles() :
            nb1+= t.count()
        for grp in self._agents :
            nb2+= len(grp)
        assert nb1 == nb2
        return nb1
    
    # Construction:
    def initializeLine( self, size, tileSize= 1.0, separation= 0.1, connect=True ):
        dist= tileSize+separation
        shape= Convex().initializeSquare(tileSize)
        self._tiles= [
            Tile( i+1, Point(dist*i, 0.0), shape.copy() )
            for i in range(size)
        ]
        self._size= size
        if connect :
            self.connectAllClose(1.1*separation)
        return self
    
    def initializeGrid( self, matrix, tileSize= 1.0, separation=0.1, connect=True ):
        dist= tileSize+separation
        self._tiles= []
        
        iTile= 0
        maxLine= len(matrix)-1
        for i in range( len(matrix) ) :
            for j in range( len(matrix[i]) ) :
                if matrix[i][j] >= 0 : 
                    iTile+= 1
                    tile= Tile(
                        iTile,
                        Point( dist*j, dist*(maxLine-i) ),
                        Convex().initializeSquare(tileSize),
                        matrix[i][j]
                    )
                    self._tiles.append( tile )
                    #matrix[i][j]= iTile
        
        self._size= iTile
        if connect :
            self.connectAllClose(1.1*separation)
        return self

    def initializeHexa( self, matrix, tileSize= 1.0, separation=0.1, connect=True ):
        cosPi06= math.cos(math.pi/6)
        dist= tileSize*cosPi06 + separation
        vdist= dist*cosPi06
        hdelta= dist*0.5
        self._tiles= []
        
        iTile= 0
        maxLine= len(matrix)-1
        for i in range( len(matrix) ) :
            for j in range( len(matrix[i]) ) :
                if matrix[i][j] >= 0 : 
                    iTile+= 1
                    iLine= maxLine-i
                    delta= (iLine%2) * hdelta
                    tile= Tile(
                        iTile,
                        Point( delta+dist*j, vdist*iLine ),
                        Convex().initializeRegular(tileSize, 6),
                        matrix[i][j]
                    )
                    self._tiles.append( tile )
                    #matrix[i][j]= iTile
        self._size= iTile
        if connect :
            self.connectAllClose(1.1*separation)
        return self
    
    def addTile( self, aTile ):
        self._size+= 1
        aTile.setId( self._size )
        self._tiles.append( aTile )
        return self._size
    
    def createSeveralTiles(self, convexes, matter):
        for c in convexes :
            self.createTile( c, matter)
        
    def createTile( self, aBody, matter= 0 ):
        self._size+= 1
        aShape= aBody.copy()
        position= aShape.setOnCenter()
        self._tiles.append( Tile( self._size, position, aShape, matter ) )
        return self._size
    
    def removeTile( self, iTile ):
        for t in self._tiles :
            if t.isConnecting(iTile) :
                t.disconnect(iTile)
        for jTile in range( iTile+1, self.size()+1 ) :
            self.changeTileID( jTile, jTile-1 )
        self._tiles.pop(iTile-1)
        self._size-=1
        return self
    
    def changeTileID( self, iTile, newID ):
        # Change connection :
        for t in self._tiles :
            if t.isConnecting(iTile) :
                t.disconnect(iTile)
                t.connect(newID)
        # Change iTile :
        self.tile(iTile).setId(newID)
        return self

    def setEpsilon(self, epsilon):
        self._epsilon= epsilon
        return self

    def setAgentFactory(self, agentFactory ):
        self._factory= agentFactory
        return self

    def clearAgents(self):
        for t in self.tiles() :
            t.clear()
        self._agents= [[]]
        return self

    def __addAgent( self, anAgent ):
        group= anAgent.group()
        while len(self._agents) <= group :
            self._agents.append([])
        assert anAgent.id() == len(self._agents[group])+1 
        self._agents[group].append(anAgent)
    
    def popAgentOn(self, iTile=1, group= 0 ):
        while len(self._agents) <= group :
            self._agents.append([])
        if iTile > self.size() :
            return False
        ag= self._factory( len(self._agents[group])+1, group )
        ag.setTile(iTile)
        ag.setPosition( self.tile(iTile).position().copy() )
        self.tile(iTile).append( ag )
        self.__addAgent(ag)
        return ag

    def connect(self, iFrom, iTo):
        self.tile(iFrom).connect(iTo)
        return self

    def connectAll(self, aList):
        for anElt in aList :
            self.connect( anElt[0], anElt[1] )

    def connectAllConditions(self, conditionFrom=lambda tfrom : True, conditionFromTo=lambda tfrom, tto : True, ):
        size= self.size()
        count= 0
        for i in range(1, size+1) :
            tili= self.tile(i)
            if conditionFrom( tili ) :
                for j in range(1, size+1) :
                    tilj= self.tile(j)
                    if conditionFromTo( tili, tilj ): # :
                       self.connect( i, j )
                       count+= 1
        return count

    def connectAllClose(self, distance):
        return self.connectAllConditions(
            conditionFromTo=lambda tileFrom, tileTo : tileFrom != tileTo and tileFrom.bodyDistance( tileTo ) < distance )

    # Distance :
    def computeDistances(self):
        s= self.size()
        self._distances= [ [ i for i in range(s+1) ] ]
        for i in range( 1, s+1 ) :
            dist= self.computeDistancesTo(i)
            self._distances.append( dist )

    def computeDistancesTo(self, iTile):
        # Initialize distances to 0:
        dists= [iTile] +  [0 for i in range( self.size() )]
        # Initialize step from iTile:
        ringNodes= self.adjacencies(iTile)
        ringDistance= 1
        # while theire is nodes to visit
        while len(ringNodes) > 0 :
            nextNodes= []
            # Visit all step nodes:
            for node in ringNodes :
                # Update distance information
                dists[node]= ringDistance
            for node in ringNodes :
                # Search for new tile to visit:
                neighbours= self.adjacencies(node)
                for candidate in neighbours :
                    if dists[candidate] == 0 :
                         nextNodes.append(candidate)
            # swith to the next step.
            ringNodes= nextNodes
            ringDistance+= 1
        # Correct 0 distance:
        dists[iTile]= 0
        return dists
    # Tile selection : 
    def selectId(self, tileCondition):
        selection= []
        for i in range(1, self.size()+1):
            if tileCondition( self.tile(i) ) :
                selection.append(i)
        return selection
    
    def selectIdSmallbox(self, maxSize):
        selection= []
        for i in range(1, self.size()+1):
            t= self.tile(i)
            tb= t.box()
            if (tb.width() < maxSize or tb.height() < maxSize) :
                selection.append(i)
        return selection
        
    # Tile operation :
    def mergeTilesIfPossible(self, iTile1, iTile2, maxError, maxSize):
        if iTile1 == iTile2 :
            return False
        if iTile2 < iTile1 :
            return self.mergeTilesIfPossible(iTile2, iTile1, maxError, maxSize)

        tile= self.tile(iTile1)
        newConvex= tile.body().copy()
        convex2= self.tile(iTile2).body()
        removed= newConvex.merge( convex2 )
        w, h= newConvex.box().dimention()
        if w > maxSize or h > maxSize :
            return False

        # Merge ok ?
        for p in removed :
            if newConvex.distancePoint(p) > maxError :
                return False
        
        # Do merge:
        newConvex.simplify( self.epsilon() )
        tile.setBody(newConvex)
        deadTile= self.tile(iTile2)
        # merge all agents
        for ag in deadTile.agents() :
            tile.append( ag )
        
        # merge connections
        adjs= deadTile.adjacencies()
        adjs.remove(iTile1)
        tile.connectAll( adjs )
        for t in self._tiles :
            if t != self.tile(iTile1) and t.isConnecting(iTile2) :
                t.connect(iTile1)
        
        self.removeTile(iTile2)
        return True
    
    def mergeTile(self, iTile, maxError, maxSize):
        t= self.tile(iTile)
        tb= t.box()
        neighborhood= [ (i, (tb + self.tile(i).box()).score() ) for i in t.adjacencies() ]
        neighborhood.sort(key=lambda tup: tup[1])
        for neighbor, val in neighborhood :
            if ( self.tile(neighbor).matter() == t.matter()
                and self.mergeTilesIfPossible( neighbor, iTile, maxError, maxSize )
            ) :
                return True
        return False

    def mergeAllPossible(self, maxError= False, exepectedSize= False):
        if not maxError :
            maxError= self.epsilon()
        if not exepectedSize :
            w, h= self.box().dimention()
            exepectedSize= max(w, h)
        notok= True
        count= 0
        minSize= exepectedSize*0.5
        maxSize= exepectedSize*1.5
        while notok :
            notok= False
            for iTile in self.selectIdSmallbox(minSize) :
                if self.mergeTile(iTile, maxError, maxSize) :
                    count+= 1
                    notok= True
                    break
        
        return count

    # Agent Collection:
    def allAgents( self, iGroup=0 ):
        alls= []
        for ags in self._agents:
            alls+= ags
        return alls
    
    def agents( self, iGroup=0 ):
        return self._agents[iGroup]

    def agentTiles( self, iGroup=0 ):
        return [ ag.tile() for ag in self.agents(iGroup) ]

    # Podable:
    def asPod( self, name= "Scene" ):
        return Pod().fromLists(
            [name], [], [self._epsilon],
            [ t.asPod() for t in self.tiles() ]
        )
    
    def fromPod( self, aPod ):
        self.clear()
        self._epsilon= aPod.value(1)
        allAgents= []
        for absTile in aPod.children() :
            t= Tile().fromPod( absTile, self._factory )
            self.addTile(t)
            for ag in t.agents() :
                self.__addAgent(ag)
        return self
    
    # string:
    def str(self, name="Scene"):
        eltStrs =[]
        for t in self.tiles() :
            eltStrs.append( f"- {t}" )
            for ag in t.agents() :
                eltStrs.append( f"  - {ag}" )
        return f"{name}:\n" + "\n".join( eltStrs )
    
    def __str__(self):
        return self.str()


# Basement: 
from .geometry import Point, Line, Box, Convex, Grid

# core coponents: 
from .agent import Agent
from .tile import Tile
from .scene import Scene

# rendering:
from .artist import Support, SupportSVG, SupportPNG, Artist

# interoperability:
from .interface import ros
from .interface import web



import random, hacka as hk

#from .artist import Artist
#from  ... import tiled

from .carrier import Carrier
from .world import World

class GameEngine( hk.AbsGame ) :

    # Initialization:
    def __init__( self, world, numberOfPlayers=1, numberOfCarriers= 1, tic= 10, seed=False ):
        super().__init__()
        self._seed= seed
        # GameEngine:
        assert( type(world) == World )
        self._numberOfPlayers= numberOfPlayers
        self._model= world
        self._model.clearAgents()
        self._model.computeDistances()
        self._initialTic= tic
        self._tic= 0
        # Initialize Players:
        for pId in range(1, self._numberOfPlayers+1) :
            for tileId in random.choices( range(1, self._model.size()+1 ), k = numberOfCarriers ) :
                self._model.popAgentOn( tileId, pId )
        self._scores= [ 0.0 for i in range(self._numberOfPlayers+1) ]
    
    # Game interface :
    def initialize(self):
        self._msg= [ hk.Pod("Hello player", [i]) for i in range(self._nbPlayer+1) ]
        self._tic= 0
        return hk.Pod("EchoGame", [self._nbPlayer, self._nbTics])
    
    def playerHand( self, iPlayer=1 ):
        # Engine :
        pod= hk.Pod( "State", [self._tic], self._scores )
        # Missions :
        pod.append( self._model.missionsAsPod() )
        # Mobiles :
        pod.append( self._model.carriersAsPod() )
        return pod

    def applyAction( self, action, iPlayer=1 ):
        # Interpret action string
        decompo= action.split(" ")
        iCarrier= 1
        while len(decompo) > 0 and iCarrier <= self.numberOfPlayers() :
            act= decompo.pop(0)
            if act == 'go' :
                clockDir= int(decompo.pop(0))
                self.setMoveAction(iPlayer, iCarrier, clockDir)
            elif act == 'do' :
                missionId= int(decompo.pop(0))
                self.setMissionAction(iPlayer, iCarrier, missionId)
            elif act != 'pass' :
                break
            iCarrier+=1
        return True

    def tic( self ):
        self.applyMoveActions()

    def isEnded( self ):
        # if the counter reach it final value
        return self.ticCounter() == 0

    def playerScore( self, iPlayer=1 ):
        # All players are winners.
        return self.score(iPlayer)

    def numberOfPlayers(self):
        return self._numberOfPlayers

    # Accessor :
    def world(self):
        return self._model
    
    def score(self, iPlayer):
        return self._scores[iPlayer]
    
    def numberOfCarriers(self, iPlayer=1):
        return len( self._model.agents(iPlayer) )
    
    def ticCounter(self):
        return self._tic

    # Game interface :
    def initialize(self, mission=None):
        self._tic= self._initialTic
        self._scores= [ 0.0 for i in range( self.numberOfPlayers()+1 ) ]
        self._model.clearMissions()
        if mission is None :
            self._model.addMissionAtRandom()
        else :
            self._model.addMission( mission[0], mission[1] )
        return self._model.asPod()

    def setMoveAction( self, iPlayer, iCarrier, clockDir ):
        # Security:
        if not(self._model.isAgent(iCarrier, iPlayer) and 0 <= clockDir and clockDir <= 12):
            return False
        self._model.agent(iCarrier, iPlayer).setMove(clockDir)
        return True
    
    def setMissionAction(self, iPlayer, iCarrier, iMission):
        # Security:
        model= self._model
        if not( model.isAgent(iPlayer, iCarrier) and model.isMission(iMission) ) :
            return False
        # Localvariable:
        carrier= model.agent(iCarrier, iPlayer)
        iFrom, iTo, pay, owner= model.mission(iMission).asTuple()
        # Mission start:
        if carrier.mission() == 0 : 
            if carrier.tile() == iFrom and owner == 0 :
                carrier.setMission( iMission )
                model.updateMission(iMission, iFrom, iTo, pay, iPlayer)
                return True
            return False
        # Mission end:
        if carrier.mission() == iMission and carrier.tile() == iTo : 
            carrier.setMission(0)
            self._scores[iPlayer]+= pay
            model.updateMission(iMission, 0, iTo, 0, iPlayer)
            model.addMissionAtRandom()
            return True
        return False

    def applyMoveActions(self):
        collision= 0
        # Visit all agents:
        for iPlayer in range( 1, self._numberOfPlayers+1 ):
            for mobile in self._model.agents(iPlayer) :
                if mobile.move() != 0 :
                    self._model.move( mobile.tile(), mobile.move() )
                    self._scores[iPlayer]+= -1
        # Clean moves:
        self._model.initializeMoves()
        self._tic-= 1
        return collision
    
    def applyMoveActionsWithCollide(self):
        collision= 0
        reserved= []
        blocked= []
        # Get players moves:
        moves= [[]]
        for iPlayer in range( 1, self._numberOfPlayers+1 ):
            moves.append([])
            for mobile in self._model.agents(iPlayer) :
                iFrom= mobile.tile()
                iTo= self._model.clockposition( iFrom, mobile.move() )
                moves[iPlayer].append( [iFrom, iTo] )
        
        # Start Collision: 
        for iPlayer in range( 1, self._numberOfPlayers+1 ):
            for m in moves[iPlayer] :
                if m[0] != m[1] and self._model.tile(m[1]).count() > 0 :
                    self._scores[iPlayer]+= -100
                    collision+= 1
                    m[1]= m[0]

        # Reserve and block:
        for iPlayer in range( 1, self._numberOfPlayers+1 ):
            for m in moves[iPlayer] :
                if m[0] != m[1] :
                    if m[1] in reserved :
                        if m[1] not in blocked :
                            blocked.append(m[1])
                    else:
                        reserved.append(m[1])

        # Apply non-bloked moves :
        for iPlayer in range( 1, self._numberOfPlayers+1 ):
            for m in moves[iPlayer] :
                if m[1] in blocked :
                    self._scores[iPlayer]+= -100
                    collision+= 1
                else:
                    self._model.teleport( m[0], m[1] )
        
        # Clean moves:
        self._model.initializeMoves()
        self._tic-= 1
        return collision

    
    
    def toward(self, iTile, iTarget):
        world= self.world()
        # If no need to move:
        if iTile == iTarget :
            return 0, iTile
        # Get candidates:
        clockdirs= world.clockBearing(iTile)
        nextTiles= world.adjacencies(iTile)
        selectedDir= clockdirs[0]
        selectedNext= nextTiles[0]
        # Test all candidates:
        for clock, tile in zip( clockdirs, nextTiles ) :
            if world._distances[tile][iTarget] < world._distances[selectedNext][iTarget] :
                selectedDir= clock
                selectedNext= tile
        # Return the selected candidates:
        return selectedDir, selectedNext

    def path(self, iTile, iTarget):
        clock, tile= self.toward(iTile, iTarget)
        move= [clock]
        path= [tile]
        while tile != iTarget :
            clock, tile= self.toward(tile, iTarget)
            move.append( clock )
            path.append( tile )
        return move, path
    
class GameMaster( hk.SequentialGameMaster ) :
    def __init__( self, world, numberOfPlayers=1, numberOfCarriers= 1, tic= 10, seed=False ):
        super().__init__(
            GameEngine( world, numberOfPlayers, numberOfCarriers, tic, seed ), 
            numberOfPlayers
    )



#----------------------------------------------------------------------------------------------------------#
#                                   H A C K A P Y  :  P O D
#----------------------------------------------------------------------------------------------------------#
# Pod : Piece Of Data
# A structure to generatize interactation between game actors (Master and player)
#----------------------------------------------------------------------------------------------------------#

class Podable():

    # Podable:
    def asPod(self):
        # Should return self as a Pod instance
        assert "Should be implemented" == None
        
    def fromPod(self):
        # Should rebuild self from a Pod instance
        assert "Should be implemented" == None
      
    # Tools:
    def podCopy(self):
        return type(self)().fromPod( self.asPod() )

class Pod(Podable):
    def __init__( self, aPod= None ):
        if aPod is None :
            self._words= []
            self._integers= []
            self._values= []
            self._children= []
        else :
            self._words= aPod.words()
            self._integers= aPod.integers()
            self._values= aPod.values()
            self._children= aPod.children()
    
    # Accessor:
    def words(self):
        return self._words
    def numberOfWords(self):
        return len( self.words() )
    def word(self, i=1):
        return self.words()[i-1]
    
    def integers(self):
        return self._integers
    def numberOfIntegers(self):
        return len( self.integers() )
    def integer(self, i=1):
        return self.integers()[i-1]
    
    def values(self):
        return self._values
    def numberOfValues(self):
        return len( self.values() )
    def value(self, i=1):
        return self.values()[i-1]
    
    def children(self):
        return self._children
    def numberOfChildren(self):
        return len( self.children() )
    def child(self, i=1):
        return self.children()[i-1]

    # Construction:
    def setWords( self, aList ):
        self._words= aList
        return self
    
    def setIntegers( self, aList ):
        self._integers= aList
        return self
    
    def setValues( self, aList ):
        self._values= aList
        return self
    
    def setChildren( self, aList ):
        self._children= aList
        return self
    
    # Collection:
    def clear(self):
        self._children= []
        return self
    
    def append(self, aChild):
        self._children.append( aChild )
        return self

    # Initialization:
    def fromLists(self, words= [], integers= [], values= [], children= []):
        self._words= [elt for elt in words ]
        self._integers= [elt for elt in integers ]
        self._values= [elt for elt in values ]
        self._children= [elt for elt in children ]
        return self
    
    def fromDico(self, aDico):
        return self

    # transfrom:
    def asDico(self, aDico):
        return self

    # Podable:
    def asPod( self ):
        return Pod().fromLists(
            self.words(),
            self.integers(),
            self.values(),
            [ child.asPod() for child in self.children() ]
        )
    
    def fromPod( self, aPod ):
        self._words= aPod.words()
        self._integers= aPod.integers()
        self._values= aPod.values()
        self._children= [ Pod().fromPod(child) for child in aPod.children() ]
        return self

    # Comparison:
    def __eq__(self, another):
        return (
            self._words == another.words()
            and self._integers == another.integers()
            and self._values == another.values()
            and self._children == another.children()
        )
    
    # String :
    def __str__(self):
        return self.str(0)

    def str(self, ident=0):
        # Get pod info
        words= self.words()
        flags= self.integers()
        values= self.values()

        # Print
        if len( words ) == 1 :
            msg= words[0] +":"
        elif len( words ) > 1 :
            msg= words[0] +": " + ", ".join( words[1:] )
        else :
            msg= "Pod:"
        if len( flags ) > 0 :
            msg+= ' ['+ ', '.join( str(i) for i in flags ) + "]"
        if len( values ) > 0 :
            msg+= ' ['+ ', '.join( str(i) for i in values ) + "]"
        # Print children
        msg+= self.strChildren( ident )
        return msg

    def strChildren( self, ident ):
        msg= ""
        newLine= '\n'
        for i in range(ident) :
            newLine+= '  '
        newLine+= '- '
        
        for c in self.children() :
            msg+= newLine + c.str(ident+1)
        return msg



import math
from .basic import Point

class Box(): # AABB Box

    # Initialization Destruction:
    def __init__( self, listOfPoint=[Point()] ):
        self.fromList( listOfPoint )

    def copy(self):
        cpy= Box()
        cpy._min= self._min.copy()
        cpy._max= self._max.copy()
        return cpy
     
    # Initialization:
    def fromList( self, listOfPoint= [Point()] ):
        self._min= Point( listOfPoint[0].x(), listOfPoint[0].y() )
        self._max= Point( listOfPoint[0].x(), listOfPoint[0].y() )
        for p in listOfPoint :
            if p.x() < self._min.x() :
                self._min.setx( p.x() )
            if p.y() < self._min.y() :
                self._min.sety( p.y() )
            if p.x() > self._max.x() :
                self._max.setx( p.x() )
            if p.y() > self._max.y() :
                self._max.sety( p.y() )
        return self

    # Accessors
    def leftFloor(self):
        return self._min
    
    def rightCeiling(self):
        return self._max
    
    def width(self) :
        return self._max.x() - self._min.x()
    
    def height(self) :
        return self._max.y() - self._min.y()
    
    def dimention(self):
        return ( self.width(), self.height() )
    
    def center(self):
        v= self._max - self._min
        return self._min + Point( v.x()*0.5, v.y()*0.5 )
    
    def perimeter(self):
        return self.width()*2 + self.height()*2
    
    def score(self):
        w, h= self.width(), self.height()
        return (self.width() + self.height()) * 1 + abs(w-h)
    
    # Construction:
    def round(self, precition=0):
        self._min.round(precition)
        self._max.round(precition)
        return self
    
    def move(self, aPoint):
        self._min+= aPoint
        self._max+= aPoint
        return self
    
    def merge( self, another ):
        self._min._x= min( self._min._x, another._min._x )
        self._min._y= min( self._min._y, another._min._y )
        self._max._x= max( self._max._x, another._max._x )
        self._max._y= max( self._max._y, another._max._y )
        return self
    
    # Transform:
    def asList(self):
        return self._min.asList() + self._max.asList()
    
    def asZip(self):
        return [self._min.asTuple(), self._max.asTuple()]
    
    # Comparison:
    def __eq__(self, another):
        return ( self._min == another.leftFloor()
            and self._max == another.rightCeiling()
        )

    # Collision:
    def isColliding(self, another):
        # no separation in anny axis
        return ( not( 
                (self._min._x <= another._min._x and self._max._x <= another._min._x )
                or (self._min._y <= another._min._y and self._max._y <= another._min._y )
                or (self._min._x > another._min._x and another._max._x < self._min._x )
                or (self._min._y > another._min._y and another._max._y < self._min._y )
            )
        )

    def isIncluding(self, another):
        # both coordinates are included
        return ( 
            ( self._min._x <= another._min._x and another._min._x < another._max._x )
            and ( self._min._y <= another._min._y and another._min._y < another._max._y )
            and ( self._min._x < another._max._x and another._max._x <= another._max._x )
            and ( self._min._y < another._max._y and another._max._y <= another._max._y )
        )
    
    # operator:
    def __add__(self, another):
        b= self.copy()
        b.merge(another)
        return b

    # to str
    def str(self): 
        # Myself :
        s= f"⌊{self._min}, {self._max}⌉"
        return s
    
    def __str__(self): 
        return self.str()


import time
from hacka import Player
from .world import World

class BasicBot( Player ):

    # Constructor:
    def __init__(self):
        super().__init__()
        self._model= World()
        self._id= 0
        self._sumResult= 0.0
        self._countResult= 0
        self._tic= 0

    # Accessor:
    def model(self):
        return self._model
    
    def ticCounter(self):
        return self._tic

    def playerId(self):
        return self._id
    
    def averageResult(self):
        return (self._sumResult/self._countResult)

    # Construction
    def resetResult(self):
        self._sumResult= 0.0
        self._countResult= 0
        return self

    # Player interface :
    def wakeUp(self, playerId, numberOfPlayers, gamePod):
        # Initialize from gamePod:
        self._model.fromPod(gamePod)
        self._id= playerId

    def perceive(self, podState):
        # update the game state:
        self._tic= self._model.setOnPodState(podState)

    def decide(self):
        return "pass"
    
    def sleep(self, result):
        self._sumResult+= result
        self._countResult+= 1
    
class ShellPlayer( BasicBot ):
    def __init__(self):
        super().__init__()
        self._action= "pass"
    
    # Player interface :
    def wakeUp(self, playerId, numberOfPlayers, gamePod):
        super().wakeUp(playerId, numberOfPlayers, gamePod)
        self._action= "pass"
        self.model().render()
        print( f"Output image : ./shot-pickndel.png" )
        print( "New game..." )
        print( "Possible Actions:" )
        print( "   - do missionId" )
        print( "   - go clockDirection" )
        print( "   - pass" )
        print( "   - stop (will 'pass' until the game ends)" )

    def perceive(self, podState):
        super().perceive(podState)
        self.model().render()

    def decide(self):
        if self._action != "stop" :
            msg= f'tic-{ self.ticCounter() }'
            msg+= f', tile-{ self._model.carrierTile() }'
            msg+= f', goal-{ self._model.carrierGoal() }'
            msg+= '\n> Enter your action: '
            self._action = input(msg)
        if self._action == "stop" :
            return "pass"
        return self._action

    def sleep(self, result):
        super().sleep(result)
        print( f"End on result: {result}" )




# Color transform:

def color( red, green, blue ):
    if red > 0xFF :
        red= 0xFF
    if green > 0xFF :
        green= 0xFF
    if blue > 0xFF :
        blue= 0xFF
    return (red<<16) + (green<<8) + (blue)

def colorFromWeb( webColor ):
    return int( webColor[1:], base=16)

def rgbColor( color ):
    b= (color & 0xFF)
    var= color >> 8
    g= (var & 0xFF)
    var= var >> 8
    r= (var & 0xFF)
    return (r, g, b)

def percentColor( color ):
    r, g, b= rgbColor(color)
    ratio= 1.0/0xff
    return round(r*ratio, 4), round(g*ratio, 4), round(b*ratio, 4)

def webColor( color ):
    string= '#'
    for c in rgbColor( color ) :
        string+= hex( (c>>4)&0xF )[2]
        string+= hex( c&0xF )[2]
    return string

def colorRatio( color, ratio ):
    r,g,b= rgbColor( color )
    r= int( r*ratio )
    g= int( r*ratio )
    b= int( r*ratio )
    return color
 


#----------------------------------------------------------#
# 
# 
#----------------------------------------------------------#
from ..geometry import Point, Grid
import yaml, cairo

def loadGridMap( path, file ) :
    if path == "" :
        path= "."
    gridmap= GridMap().load( path, file )
    return gridmap.asTllGrid()

class GridMap :
    # Initialization
    def __init__(self, position= (0.0, 0.0), resolution=0.5):
        self._grid= [[0]]
        self._position= position
        self._resolution= resolution
        self._occupied= 0.8
        self._free= 0.2

    # Accessors
    def grid(self):
        return self._grid
    
    def dimention(self):
        return ( len(self._grid[0]), len(self._grid) )
    
    def position(self):
        return self._position

    def resolution(self):
        return self._resolution

    def occupiedTreshold(self):
        return self._occupied        
    
    def freeTreshold(self):
        return self._free

    # File managment
    def load( self, path, file ):
        config= {}
        with open( path+"/"+file, "r") as file:
            # Charger le contenu du fichier en tant que dictionnaire Python
            config = yaml.safe_load( file )

        assert config['mode'] == "trinary"
        assert config['negate'] in [0, False]

        image = cairo.ImageSurface.create_from_png( path+"/"+ config['image'] )
        
        origin= config['origin']
        self._position= (origin[0], origin[1])
        self._resolution= config['resolution']
        self._occupied= config['occupied_thresh']
        self._free= config['free_thresh']
    
        # build the grid :
        assert image.get_format() == cairo.Format.RGB24
        height= image.get_height()
        stride= image.get_stride()
        memoryview = image.get_data()

        # build the grid
        self._grid= []
        for iLine in range(height) :
            self._grid.append( [] )
            for iCol in range(0, stride, 4) :
                i= iLine*stride+iCol
                pix= memoryview[i] + memoryview[i+1] + memoryview[i+2]
                pix/= 3
                status= round( (255.0 - pix) / 255.0, 3 )
                self._grid[iLine].append( status )
        # out
        return self
    
    def cellIsFree(self, x, y):
        return (self.cell(x, y) <= self._free)

    def cellIsOccupied(self, x, y):
        return (self.cell(x, y)  >= self._occupied)

    def cellIsUncertain(self, x, y):
        return (not self.cellIsFree(x, y) and not self.cellIsOccupied(x,y))
    
    # Morphing
    def asStatesGrid(self):
        # Build a [Free(0), Occupied(1), Unknown(2)] map
        width, height = self.dimention()
        fouMap= []
        for iLine in range(height) :
            fouMap.append( [] )
            for iCol in range(width) :
                status= self._grid[iLine][iCol]
                if status >= self._occupied :
                    fouMap[iLine].append( 1 )
                elif status <= self._free :
                    fouMap[iLine].append( 0 )
                else :
                    fouMap[iLine].append( 2 )
        return fouMap
    
    def asTllGrid(self):
        px, py= self.position()
        d= self.resolution() * 0.5
        return Grid(
            self.asStatesGrid(),
            Point(px-d, py-d),
            self.resolution()
        )


from .color import color, colorRatio, colorRatio, rgbColor, percentColor, webColor, colorFromWeb
from .support import AbsSupport, Support, SupportSVG
from .supportCairo import SupportPNG
from ..scene import Scene 

def drawScene(scene, filePath= "shot-tiled.png", width= 1600, height= 1200):
    pablo= Artist().initializePNG( filePath, width, height )
    pablo.fit(scene)
    pablo.drawScene(scene)
    pablo.flip()

def drawConvexes(convexes, filePath= "shot-tiled.png", width= 1600, height= 1200):
    scene= Scene(convexes)
    drawScene(scene,filePath, width, height) 

# Artist:
class Brush():
    def __init__(self, fill= 0xff6644, stroke= 0x991100, width= 4 ):
        self.fill= fill
        self.stroke= stroke
        self.width= width

class Artist():
    def __init__(self):
        #  Initialize support:
        self._support= Support()

        # Initialize brush :
        self._background= 0xb86e00
        self._panel= [
            Brush(0xffcd80, 0x603800, 4), # 0-Free
            Brush(0xff6644, 0x991100, 4), # 1-Red
            Brush(0x70f050, 0x20770a, 4), # 2-Green
            Brush(0x6666ff, 0x1111aa, 4), # 3-Blue
            Brush(0xfd9622, 0xdd550a, 4), # 4-Orange
            Brush(0xdd77ff, 0x8800aa, 4), # 5-Purple
            Brush(0x66ddee, 0x117799, 4), # 6-Cian
            Brush(0xffffff, 0xbbbbbb, 4), # 7-White
            Brush(0x888888, 0x555555, 4), # 8-Grey
            Brush(0x444444, 0x000000, 4), # 9-Black

            Brush(0x603800, 0xffcd80, 4), # 10-Background
            Brush(0x991100, 0xff6644, 4), # 11-Red
            Brush(0x20770a, 0x70f050, 4), # 12-Green
            Brush(0x1111aa, 0x6666ff, 4), # 13-Blue
            Brush(0xdd550a, 0xfd9622, 4), # 14-Orange
            Brush(0x8800aa, 0xdd77ff, 4), # 15-Purple
            Brush(0x117799, 0x66ddee, 4), # 16-Cian
            Brush(0xbbbbbb, 0xffffff, 4), # 17-White
            Brush(0x555555, 0x888888, 4), # 18-Grey
            Brush(0x000000, 0x444444, 4)  # 19-Black
        ]
        self._fontSize= 16

        # Initialize Frame :
        self._x= 0.0
        self._y= 0.0
        self._scale= 100.0
        self.flip()

    # Construction:
    def initializeSVG(self, filePath, width= 1600, height= 1200):
        self._support= SupportSVG( width, height, filePath )
        self.flip()
        return self
    
    def initializePNG(self, filePath, width= 1600, height= 1200):
        self._support= SupportPNG( width, height, filePath )
        self.flip()
        return self

    # Accessor:
    def support(self):
        return self._support

    def content(self):
        return self._support.content()

    def clear(self):
        h= self._support.height()
        w= self._support.width()
        self._support.clear()
        self._support.fillPolygon(
            [0, 0, w, w],
            [0, h, h, 0],
            self._background
        )
        return self
    
    def render( self ):
        return self._support.render()

    def camera( self ):
        return (self._x, self._y)
        
    def scale( self ):
        return (self._scale)
    
    # Setters:
    def setCamera( self, x, y ):
        self._x, self._y= x, y
        return self
        
    def setScale( self, scale ):
        self._scale= scale
        return self
    
    def fit(self, anObj):
        return self.fitBox( anObj.box() )

    def fitBox( self, aBox, marge=10 ):
        marge= marge*2
        minx, miny= aBox.leftFloor().asTuple()
        maxx, maxy= aBox.rightCeiling().asTuple()
        self.setCamera( (minx+maxx)*0.5, (miny+maxy)*0.5 )
        distx= maxx-minx
        disty= maxy-miny
        ratioX, ratioY= 1.0, 1.0
        if distx != 0.0 :
           ratioX= (self._support.width()-marge)/distx
        if disty != 0.0 :
            ratioY= (self._support.height()-marge)/disty
        self.setScale( min(ratioX, ratioY) )
        return self

    # Panel managments:

    # Transformation World <-> Frame
    def toFrame(self, x, y ):
        dwidth= self._support.width()*0.5
        dheight= self._support.height()*0.5
        dx= (x-self._x)*self._scale
        dy= (y-self._y)*-self._scale
        return dx+dwidth, dy+dheight

    def xToFrame(self, x ):
        dwidth= self._support.width()*0.5
        return (x-self._x)*self._scale + dwidth

    def yToFrame(self, y ):
        dheight= self._support.height()*0.5
        return (y-self._y)*-self._scale + dheight

    def toWorld(self, pixx, pixy):
        return 0, 0

    # Drawing primitives:
    def tracePoint( self, x, y, brush= Brush() ):
        pixx, pixy= self.toFrame( x, y )
        self._support.fillCircle( pixx, pixy, brush.width, brush.stroke )
        return self

    def traceLine( self, ax, ay, bx, by, brush= Brush() ):
        pixxA, pixyA= self.toFrame( ax, ay )
        pixxB, pixyB= self.toFrame( bx, by )
        self._support.traceLine( pixxA, pixyA, pixxB, pixyB, brush.stroke, brush.width)
        return self

    def traceCircle( self, x, y, radius, brush= Brush()):
        pixx, pixy= self.toFrame(x, y)
        self._support.traceCircle( pixx, pixy, radius*self._scale, brush.stroke, brush.width)
        return self

    def fillCircle( self, x, y, radius, brush= Brush()):
        pixx, pixy= self.toFrame(x, y)
        self._support.fillCircle( pixx, pixy, radius*self._scale, brush.fill )
        return self

    def drawCircle( self, x, y, radius, brush= Brush() ):
        pixx, pixy= self.toFrame( x, y )
        self._support.drawCircle( pixx, pixy, radius*self._scale, brush.fill, brush.stroke, brush.width )
        return self

    def tracePolygon( self, coordXs, coordYs, brush= Brush() ):
        self._support.tracePolygon(
            [ self.xToFrame( x ) for x in coordXs ],
            [ self.yToFrame( y ) for y in coordYs ],
            brush.stroke, brush.width
        )
        return self

    def fillPolygon( self, coordXs, coordYs, brush= Brush() ):
        self._support.fillPolygon(
            [ self.xToFrame( x ) for x in coordXs ],
            [ self.yToFrame( y ) for y in coordYs ],
            brush.fill
        )
        return self

    def drawPolygon( self, coordXs, coordYs, brush= Brush() ):
        self._support.drawPolygon(
            [ self.xToFrame( x ) for x in coordXs ],
            [ self.yToFrame( y ) for y in coordYs ],
            brush.fill, brush.stroke, brush.width
        )
        return self

    # Writting primitives:
    def write( self, x, y, text, brush= Brush() ):
        self._support.write(
            self.xToFrame(x), self.yToFrame(y),
            text, brush.stroke, self._fontSize
        )
        return self

    # Drawing frame:
    def drawFrameGrid( self, step= 1.0, color=None ):
        
        pixX, pixY= self.toFrame( 0, 0 )
        pixStep= step*self._scale

        while pixX > pixStep :
            pixX-= pixStep
        
        while pixY > pixStep :
            pixY-= pixStep
        
        width= self._support.width()
        height= self._support.height()

        if not color :
            color= self._panel[0].stroke
        
        # Vertical
        for i in range( (int)(width/pixStep)+1 ) :
            self._support.traceLine( pixX+(pixStep*i), 10, pixX+(pixStep*i), height-10, color, 2 )
        # Horizontal
        for i in range( (int)(height/pixStep)+1 ) :
            self._support.traceLine( 10, pixY+(pixStep*i), width-10, pixY+(pixStep*i), color, 2 )
        return self

    def drawFrameAxes( self ):
        brush= Brush( 0xE26060,  0xE26060, 6 )
        self.traceLine(  0, 0, 1, 0, brush )
        brush.stroke= 0x60E260
        self.traceLine(  0, 0, 0, 1, brush )
        brush.stroke= 0x0606E2
        self.tracePoint( 0, 0, brush )
        return self
    
    # Drawing map:
    def drawConvex( self, shape, brushId=0, px=0, py=0 ):
        listxs, listys= shape.asLists(px, py)
        self.drawPolygon( listxs, listys, self._panel[ brushId%len(self._panel) ] )
    
    def fillConvex( self, shape, brushId=0, px=0, py=0 ):
        listxs, listys= shape.asLists(px, py)
        self.fillPolygon( listxs, listys, self._panel[ brushId%len(self._panel) ] )
    
    def drawTile( self, aTile ):
        env= aTile.body().asZipped()
        self.drawPolygon(
            [p[0] for p in env],
            [p[1] for p in env],
            self._panel[ aTile.matter() ]
        )
    
    def writeTile( self, aTile ):
        minx, miny= aTile.box().leftFloor().asTuple()
        x, y= aTile.position().asTuple()
        x= x+(minx-x)*2/3
        y= y+(miny-y)*2/3
        self.write( x, y, str(aTile.id()), self._panel[ aTile.matter() ] )

    def drawAgent( self, agent, brushId ):
        self.fillConvex(
            agent.body(),
            brushId )
        minx, miny= agent.box().leftFloor().asTuple()
        x, y= agent.position().asTuple()
        self.write( x, y, str(agent.id()), self._panel[brushId] )
    
    def drawSceneNetwork( self, aScene ):
        for tile in aScene.tiles() :
            cx, cy= tile.position().asTuple()
            self.tracePoint( cx, cy, self._panel[ tile.matter() ] )

        for fromId, toId in aScene.edges() :
            fromX, fromY= aScene.tile( fromId ).position().asTuple()
            brush= self._panel[ aScene.tile( fromId ).matter() ]
            toX, toY= aScene.tile( toId ).position().asTuple()
            self.traceLine( fromX, fromY, toX, toY, brush )

    def drawSceneTiles( self, aScene ):
        for tile in aScene.tiles() :
            self.drawTile( tile )

    def drawSceneAgents( self, aScene ):
        for tile in aScene.tiles() :
            x, y= tile.position().asTuple()
            position= (x+0.1, y+0.1)
            for agent in tile.agents() :
                self.drawAgent( agent, agent.matter() )
    
    def writeSceneTiles( self, aScene ):
        for tile in aScene.tiles() :
            self.writeTile( tile )

    def drawScene( self, aScene ):
        self.drawSceneNetwork(aScene)
        self.drawSceneTiles(aScene)
        self.writeSceneTiles(aScene)
        self.drawSceneAgents(aScene)

    # Control:
    def flip(self):
        ref= self._support.flip()
        self.clear()
        return ref



from .color import webColor

class AbsSupport():

    # Accessor:
    def width(self):
        return 1200
    
    def height(self):
        return 800
    
    # Control:
    def clear(self):
        pass
    
    def render( self ):
        return None
    
    def flip( self ):
        rendering= self.render()
        self.clear()
        return rendering
    
    # Drawing primitives:
    def traceLine( self, pixxA, pixyA, pixxB, pixyB, strokeColor, strokeWidth ):
        pass

    def traceCircle( self, pixx, pixy, radius, strokeColor, strokeWidth ):
        pass

    def fillCircle( self, pixx, pixy, radius, fillColor ):
        pass

    def drawCircle( self, pixx, pixy, radius, fillColor, strokeColor, strokeWidth):
        pass

    def tracePolygon( self, pixXs, pixYs, strokeColor, strokeWidth ):
        pass

    def fillPolygon( self, pixXs, pixYs, fillColor ):
        pass

    def drawPolygon( self, pixXs, pixYs, fillColor, strokeColor, strokeWidth ):
        pass

    # Writting primitives:
    def write( self, pixXs, pixYs, text, color, fontSize ):
        pass

class Support( AbsSupport ):

    def __init__(self, width= 800, height= 600 ):
        self._width= width
        self._height= height
        self._canvas= []
    
    # Accessor: 
    def width(self):
        return self._width
    
    def height(self):
        return self._height
    
    def canvas(self):
        return self._canvas
    
    # Control:
    def clear(self):
        self._canvas= []
    
    def content(self):
        return '\n'.join(self._canvas) 

    def render(self):
        return f'<svg width="{self.width()}" height="{self.height()}">\n' + '\n'.join( self._canvas + ['</svg>'] ) 

    def flip(self):
        arts= self.render()
        self.clear()
        return arts

    def save( self, filePath ):
        file = open( filePath, "w")
        file.write( '<?xml version="1.0" encoding="UTF-8"?>\n' )
        arts= self.render()
        file.write( arts )
        file.close()
        return arts

    # Drawing primitives:
    def traceLine( self, pixxA, pixyA, pixxB, pixyB, strokeColor, strokeWidth ):
        self._canvas.append( f'<line x1="{pixxA}" y1="{pixyA}" x2="{pixxB}" y2="{pixyB}" style="stroke:{webColor(strokeColor)};stroke-width:{strokeWidth}"/>' )
        return self

    def traceCircle( self, pixx, pixy, radius, strokeColor, strokeWidth ):
        self._canvas.append( f'<circle r="{radius}" cx="{pixx}" cy="{pixy}" fill="none" stroke="{webColor(strokeColor)}" stroke-width="{strokeWidth}" />' )
        return self

    def fillCircle( self, pixx, pixy, radius, fillColor ):
        self._canvas.append( f'<circle r="{radius}" cx="{pixx}" cy="{pixy}" fill="{webColor(fillColor)}" />' )
        return self

    def drawCircle( self, pixx, pixy, radius, fillColor, strokeColor, strokeWidth):
        self._canvas.append( f'<circle r="{radius}" cx="{pixx}" cy="{pixy}" fill="{webColor(fillColor)}" stroke="{webColor(strokeColor)}" stroke-width="{strokeWidth}" />' )
        return self
    
    def tracePolygon( self, pixXs, pixYs, strokeColor, strokeWidth ):
        points= " ".join( [ f'{x},{y}' for x, y in zip(pixXs, pixYs) ] )
        self._canvas.append( f'<polygon points="{points}" style="fill:none;stroke:{webColor(strokeColor)};stroke-width:{strokeWidth}" />' )
        return self

    def fillPolygon( self, pixXs, pixYs, fillColor ):
        points= " ".join( [ f'{x},{y}' for x, y in zip(pixXs, pixYs) ] )
        self._canvas.append( f'<polygon points="{points}" fill="{webColor(fillColor)}" />' )
        return self

    def drawPolygon( self, pixXs, pixYs, fillColor, strokeColor, strokeWidth ):
        points= " ".join( [ f'{x},{y}' for x, y in zip(pixXs, pixYs) ] )
        self._canvas.append( f'<polygon points="{points}" style="fill:{webColor(fillColor)};stroke:{webColor(strokeColor)};stroke-width:{strokeWidth}" />' )
        return self
    
    # Writting primitives:
    def write( self, pixXs, pixYs, text, color, fontSize ):
        self._canvas.append( f'<text x="{pixXs}" y="{pixYs}" fill="{webColor(color)}" font-family="Verdana" font-size="{fontSize}">{text}</text>" />' )
        return self

class SupportSVG( Support ):
    
    def __init__(self, width= 800, height= 600, filePath= "shot-hacka.svg" ):
        super().__init__(width, height)
        self._filePath= filePath
        self.save( self._filePath )
    
    # Accessor: 
    def filePath(self):
        return self._filePath
    
    # Control:

    def flip(self):
        arts= self.save( self._filePath )
        self.clear()
        return arts

    def save( self, filePath ):
        file = open( filePath, "w")
        file.write( '<?xml version="1.0" encoding="UTF-8"?>\n' )
        arts= self.render()
        file.write( arts )
        file.close()
        return arts
    
